#!/usr/bin/perl
# From: https://github.com/odrevet/MPO, Olivier Drevet, odrevet@gmail.com, http://olivier.drevet.free.fr, Lyon, France
# Modified October 2018 by Herb Weiner <herbw@wiskit.com> to (1) operate correctly under Windows & (2) remove tmp file

use strict; use warnings;
use Getopt::Long ;
use Image::MetaData::JPEG;
use Image::MetaData::JPEG::Record;
use Image::MetaData::JPEG::data::Tables
    qw(:JPEGgrammar :Endianness :RecordTypes);

#all references in parenthesis are refering to the CIPA DC-007-2009 manual
#according chapter and paragraph. 

#get the filenames from command line arguments

my $filename_left;
my $filename_right;

########################################
#Default values

#display help
my $help = 0;

#output default name
my $filename_out = "out.MPO";

#distance in milimeters between the two viewpoints
my $baseline_length = 77;

########################################
#Read command line arguments

GetOptions (
    'left|l:s'     => \$filename_left, 
    'right|r:s'    => \$filename_right, 
    'out|o:s'      => \$filename_out,
    'baseline|b:i' => \$baseline_length,
    'help|h'       => \$help
    );

if($help or 
   !defined $filename_left or 
   !defined $filename_right){
    usage();
    exit;
}

#Add the APP2 segment in left and right, with optional values
#save left as $filename_out and right as tmp
add_APP2($filename_left, $filename_right, $filename_out, $baseline_length);

#concat the two modified jpg files into the MPO
concatfiles($filename_out, 'tmp');

unlink ('tmp')								#	herbw@wiskit.com
	or die "Couldn't remove file 'tmp' $!";	#	herbw@wiskit.com

#######################################
sub add_APP2{
    my ($filename_left, $filename_right) = @_;
    
    #Hard coded values
    #The size in decimal of the APP2 segment to be written
    #those values must be modified manually when adding/removing a record
    my $APP2_size_left = 158 + 2; # +2 Bytes for the APP Marker
    my $APP2_size_right = 96 + 2;

    #Number of images to be packed in the MPO file
    my $NUMBER_OF_IMAGES = 2;

    ########################################
    #get the size in Byte of each jpg file in order to set the MPEntries
    #file size values
    my $file_size_left = (stat($filename_left))[7];
    my $file_size_right = (stat($filename_right))[7];

    ########################################
    #read the image filname from the command line arguments
    my $file_left = new Image::MetaData::JPEG($filename_left, 1, 'NOPARSE');
    my $file_right = new Image::MetaData::JPEG($filename_right, 1, 'NOPARSE');

    #create the APP2 segment, do not parse it as it is empty
    my $buffer_left = '';  #the buffer that will contains the data to write
    my $segref_left = new Image::MetaData::JPEG::Segment('APP2', 
                                                         \$buffer_left, 
                                                         'NOPARSE');

    my $buffer_right;  #the buffer that will contains the data to write
    my $segref_right = new Image::MetaData::JPEG::Segment('APP2', 
                                                          \$buffer_right, 
                                                          'NOPARSE');

    #insert the APP2 segment into the image file
    $file_left->insert_segments($segref_left);
    $file_right->insert_segments($segref_right);

    ########################################################
    #MP EXTENSION  (5.2)

    #the APP2 MARKER PREFIX and FIELD LENGTH are automatically created
    #by the Image::MetaData::JPEG module

    #MP FORMAT IDENTIFIER (5.2.1)
    #A Null-Terminated Identifier in ASCII: MPF\0

    my $MP_FORMAT_IDENTIFIER_SIZE = 4;
    my $MP_FORMAT_IDENTIFIER = 
        chr(hex '4D').
        chr(hex '50').
        chr(hex '46').
        chr(hex '00');

    my $recref_mp_fid = $segref_left->create_record('MP_FORMAT_IDENTIFIER', 
                                                    $BYTE, 
                                                    \$MP_FORMAT_IDENTIFIER, 
                                                    $MP_FORMAT_IDENTIFIER_SIZE);

    #save the first file to disk with the FID
    $buffer_left  = $recref_mp_fid->get();
    $file_left->save($filename_out);

    #Find the FID offset from the begginning of the file
    open(FH_LEFT, '<', $filename_out) or die $!;
    binmode (FH_LEFT);								#	herbw@wiskit.com

    local $/ = "MPF\0";  #change the endline with the FID
    my @a = <FH_LEFT>;
    my $FID_OFFSET = length $a[0];

    #the endianess tag follow the FID offset
    my $ENDIANESS_TAG_OFFSET = $FID_OFFSET;

    ## MP HEADER (5.2.2)
    #the MP HEADER is composed of the MP_ENDIAN and the OFFSET_TO_FIRST_IFD

    #MP_ENDIAN (5.2.2.1)
    #we are using LITTLE ENDIANESS: Less Significative Bits first
    my $MP_ENDIAN = 
        chr(hex '49').
        chr(hex '49').
        chr(hex '2A').
        chr('00');
    my $recref_mp_endian = $segref_left->create_record('MP_ENDIAN', 
                                                       $BYTE,
                                                       \$MP_ENDIAN,
                                                       4);
    #OFFSET_TO_FIRST_IFD (5.2.2.2)
    #offset of the first IFD. It is at the next Byte (\0x08)
    my $OFFSET_TO_FIRST_IFD = 
        chr(hex '08').
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');
    
    my $recref_offset_to_first_ifd = $segref_left->create_record('OFFSET_TO_FIRST_IDF', 
                                                                 $BYTE,
                                                                 \$OFFSET_TO_FIRST_IFD,
                                                                 4);

    ###############################
    ##Start of MP INDEX IFD (5.2.3)
    #for the first individual image only. Each field is introduced by a tag. 
    #count the number of fields to be declared (Version, Number Of Images, MP Entry)
    my $MPI_COUNT = 
        chr(hex '03').
        chr(hex '00');
    my $recref_mpi_count = $segref_left->create_record('MPI_COUNT', 
                                                       $BYTE,
                                                       \$MPI_COUNT,
                                                       2);
    #Version
    my $MPI_VERSION =
        chr(hex '00'). #Tag
        chr(hex 'b0').
        chr(hex '07'). #Type (undefined)
        chr(hex '00').
        chr(hex '04'). #Length of 4 ASCII CHARS
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '30'). #Version Number 0100 in ASCII
        chr(hex '31').
        chr(hex '30').
        chr(hex '30');

    my $recref_mpi_version = $segref_left->create_record('MPI_VERSION', 
                                                         $BYTE,
                                                         \$MPI_VERSION,
                                                         12);

    #NUMBER OF IMAGES (5.2.3.2)
    my $hex_number_of_images = sprintf('%x', $NUMBER_OF_IMAGES);
    my $MPI_NUMBER_OF_IMAGES =
        chr(hex '01'). #Tag
        chr(hex 'b0').
        chr(hex '04'). #Type: Long
        chr(hex '00').
        chr(hex '01'). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex $hex_number_of_images). #Number of images
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_mpi_number_of_images = $segref_left->create_record(
        'MPI_NUMBER_OF_IMAGES', 
        $BYTE,
        \$MPI_NUMBER_OF_IMAGES,
        12
        );

    ##################################
    #OFFSET Of MP Entries vales
    #Endianess             4 Bytes
    #Offset of first IFD   4 Bytes
    #MP Index Count        2 Bytes
    #Version              12 Bytes
    #Number of Image      12 Bytes
    #MP Entry             12 Bytes
    #Offset of next IFD    4 Bytes
    #                    ----------
    #TOTAL                50 Bytes
    #In Hexadecimal     0x32 Bytes
    my $OFFSET_LENGTH_DEC = 4 + 4 + 2 + 12 + 12 + 12 + 4;
    my $OFFSET_LENGTH_HEX = sprintf("%x", $OFFSET_LENGTH_DEC);
    my $mpe_tag_count = sprintf("%x", 16 * $NUMBER_OF_IMAGES);

    my $MPE_TAG = 
        chr(hex '02'). #TAG
        chr(hex 'b0').
        chr(hex '07'). #Type
        chr(hex '00').
        chr(hex $mpe_tag_count). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex $OFFSET_LENGTH_HEX). # Offset where are the MPEntries values
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');
    my $recref_mpe_tag = $segref_left->create_record('MPE_TAG', 
                                                     $BYTE,
                                                     \$MPE_TAG,
                                                     12);
    #OFFSET OF NEXT IFD
    # Offset Details:
    # IFD of 16 Bytes per Image:  n * 16 = 32 Bytes. given n = 2 Images
    # + Offset of 50 Bytes
    # TOTAL of 82 Bytes <=> \0x52

    my $next_ifd_offset_value = 16 * $NUMBER_OF_IMAGES + $OFFSET_LENGTH_DEC;
    my $next_ifd_offset_value_hex = sprintf("%x", $next_ifd_offset_value);

    my $NEXT_IFD_OFFSET =
        chr(hex $next_ifd_offset_value_hex).
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_next_ifd_offset = $segref_left->create_record('NEXT_IFD_OFFSET', 
                                                             $BYTE,
                                                             \$NEXT_IFD_OFFSET,
                                                             4);
    #End of the MPIndex IFD
    ##########################################
    #MP ENTRY: one per image (5.2.3.3)

    #array with references to record of mpi values
    #one reference per image
    my @recref_mpi_values;

    ##################################
    # Hard coded values

    # OFFSET OF ENDIANESS TAG FROM SOI: 1C
    # SIZE OF FILE 1 = original file size + APP2 size
    # FILE 2 TO ENDIANESS OFFSET = SIZE OF FILE 1 - OFFSET OF ENDIANESS TAG FROM SOI

    my $file_size_left_dec = $file_size_left + $APP2_size_left;
    my $file_size_left_hex = sprintf('%08x', $file_size_left_dec);
    my $file_size_right_hex = sprintf('%08x', $file_size_right + $APP2_size_right);
    my $file_data_offset_hex = sprintf('%08x', 
                                       $file_size_left_dec - 
                                       $ENDIANESS_TAG_OFFSET);

    my @file_size_chunk;  #explode the file size in Byte to write in the segment
    my $step=2;           #2 hexadecimal characters per Byte

    for(my $i=$step;$i<=$step*4;$i+=$step){
        push (@file_size_chunk, substr $file_size_left_hex, -$i, $step);
    }

    my $MPI_VALUES =     #Individual Image Attributes (5.2.3.3.1) (Figure 8)
        chr(hex '02').   ##Type Code (24 bits) (Table 4) (MultiFrameDisparity)
        chr(hex '00').
        chr(hex '02').
        chr(0b10000000). ##3bits:Image Date format, 2 bits:reserved, 3 bits:flags 
        chr(hex $file_size_chunk[0]).  #Individual Image Size (5.2.3.3.2) Big Endian
        chr(hex $file_size_chunk[1]). 
        chr(hex $file_size_chunk[2]). 
        chr(hex $file_size_chunk[3]). 
        chr(hex '00').   #Individual Image Data Offset (5.2.3.3.3) Must be NULL
        chr(hex '00').
        chr(hex '00'). 
        chr(hex '00'). 
        chr(hex '00').   #Independent Image Entry Number 1 (5.2.3.3.4)
        chr(hex '00').
        chr(hex '00').   #Independent Image Entry Number 2
        chr(hex '00');

    push @recref_mpi_values,  $segref_left->create_record('MPI_VALUES', 
                                                          $BYTE,
                                                          \$MPI_VALUES,
                                                          16);

    @file_size_chunk = ();  #empty the array 
    for(my $i=$step;$i<=$step*4;$i+=$step){
        push (@file_size_chunk, substr $file_size_right_hex, -$i, $step);
    }

    my @file_offset_chunk;
    for(my $i=$step;$i<=$step*4;$i+=$step){
        push (@file_offset_chunk, substr $file_data_offset_hex, -$i, $step);
    }

    my $MPI_VALUES_B =   #Individual Image Attributes (5.2.3.3.1) (Figure 8)
        chr(hex '02').   ##Type Code (24 bits) (Table 4) (MultiFrameDisparity)
        chr(hex '00').
        chr(hex '02').
        chr(0b00000000). ##3bits:Image Date format, 2 bits:reserved, 3 bits:flags 
        chr(hex $file_size_chunk[0]). #Individual Image Size (5.2.3.3.2) Big Endian
        chr(hex $file_size_chunk[1]). 
        chr(hex $file_size_chunk[2]). 
        chr(hex $file_size_chunk[3]). 
        chr(hex $file_offset_chunk[0]). #Individual Image Data Offset (5.2.3.3.3) BE
        chr(hex $file_offset_chunk[1]).
        chr(hex $file_offset_chunk[2]).
        chr(hex $file_offset_chunk[3]).
        chr(hex '00').   #Independent Image Entry Number 1 (5.2.3.3.4)
        chr(hex '00').
        chr(hex '00').   #Independent Image Entry Number 2
        chr(hex '00');

    push @recref_mpi_values, $segref_left->create_record('MPI_VALUES', 
                                                         $BYTE,
                                                         \$MPI_VALUES_B,
                                                         16);

    ##########################################
    #Start of MPAttributes IFD (5.2.4)
    #count the number of fields to be declared
    my $MPA_COUNT = 
        chr(hex '04').
        chr(hex '00');
    my $recref_mpa_count = $segref_left->create_record('MPA_COUNT', 
                                                       $BYTE,
                                                       \$MPA_COUNT,
                                                       2);

    #MP Individual Image Number (5.2.4.2)
    my $MPA_INDIVIDUAL_IMAGE_NUMBER =
        chr(hex '01'). #Tag
        chr(hex 'b1').
        chr(hex '04'). #Type
        chr(hex '00').
        chr(hex '01'). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '01'). #VALUE
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_mpa_individual_image_number = $segref_left->create_record(
        'MPA_INDIVIDUAL_IMAGE_NUMBER', 
        $BYTE,
        \$MPA_INDIVIDUAL_IMAGE_NUMBER,
        12
        );

    #BASE VIEWPOINT NUMBER (5.2.4.5)
    my $MPA_BASE_VIEWPOINT_NUMBER =
        chr(hex '04'). #Tag
        chr(hex 'b2').
        chr(hex '04'). #Type
        chr(hex '00').
        chr(hex '01'). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '01'). #Value
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_mpa_base_viewpoint_number = $segref_left->create_record(
        'MPA_BASE_VIEWPOINT_NUMBER', 
        $BYTE,
        \$MPA_BASE_VIEWPOINT_NUMBER,
        12
        );

    #MPA Convergence Angle (5.2.4.6)
    my $MPA_CONVERGENCE_ANGLE =
        chr(hex '05'). #Tag
        chr(hex 'b2').
        chr(hex '0a'). #Type: SRATIONAL
        chr(hex '00').
        chr(hex '01'). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '88'). #Offset Value
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_mpa_convergence_angle = $segref_left->create_record(
        'MPA_CONVERGENCE_ANGLE', 
        $BYTE,
        \$MPA_CONVERGENCE_ANGLE,
        12
        );

    #MP Baseline Length (5.2.4.7)
    my $MPA_BASELINE_LENGTH =
        chr(hex '06'). #Tag
        chr(hex 'b2').
        chr(hex '05'). #Type: RATIONAL
        chr(hex '00').
        chr(hex '01'). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '90'). #offset value
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_mpa_baseline_length = $segref_left->create_record(
        'MPA_BASELINE_LENGTH', 
        $BYTE,
        \$MPA_BASELINE_LENGTH,
        12
        );

    #OFFSET OF NEXT IFD
    my $NEXT_IFD_OFFSET_NULL =
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_next_ifd_offset_null = $segref_left->create_record('NEXT_IFD_OFFSET', 
                                                                  $BYTE,
                                                                  \$NEXT_IFD_OFFSET_NULL,
                                                                  4);

    #convert the baseline value into hex and array of byte
    my $baseline_length_hex = sprintf("%08x", $baseline_length);
    my @baseline_length_chunk;
    for(my $i=$step;$i<=$step*4;$i+=$step){
        push (@baseline_length_chunk, substr $baseline_length_hex, -$i, $step);
    }


    #MP ATTRIBUT VALUES IFD
    my $MPA_VALUES =
        chr(hex '00'). #Convergence angle (5.2.4.6)
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '01').
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex $baseline_length_chunk[0]). #Baseline length (5.2.4.7)
        chr(hex $baseline_length_chunk[1]).
        chr(hex $baseline_length_chunk[2]).
        chr(hex $baseline_length_chunk[3]).
        chr(hex 'e8').
        chr(hex '03').
        chr(hex '00').
        chr(hex '00');

    my $recref_mpa_values = $segref_left->create_record('MPA_VALUES', 
                                                        $BYTE,
                                                        \$MPA_VALUES,
                                                        16);
    ##########################################
    #Segments for the second image
    #If a record in the  first image fits exacly for the second image, this record
    #will be used as a second image record. In other terms: only differant records
    #from the first image record will be created

    #Count the number of attribute
    #There is one attribute more than in the first image. Which is identical to
    #the MPI version tag, so we will be using the MPI version tag of the first image
    #as the MPA version tag for the second image
    my $MPA_COUNT_B = 
        chr(hex '05').
        chr(hex '00');
    my $recref_b_mpa_count = $segref_right->create_record('MPA_COUNT', 
                                                          $BYTE,
                                                          \$MPA_COUNT_B,
                                                          2);
    #MP Individual Image Number (5.2.4.2)
    my $MPA_INDIVIDUAL_IMAGE_NUMBER_B =
        chr(hex '01'). #Tag
        chr(hex 'b1').
        chr(hex '04'). #Type
        chr(hex '00').
        chr(hex '01'). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '02'). #VALUE
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_b_mpa_individual_image_number = $segref_left->create_record(
        'MPA_INDIVIDUAL_IMAGE_NUMBER', 
        $BYTE,
        \$MPA_INDIVIDUAL_IMAGE_NUMBER_B,
        12
        );

    #MPA Convergence Angle (5.2.4.6)
    my $MPA_CONVERGENCE_ANGLE_B =
        chr(hex '05'). #Tag
        chr(hex 'b2').
        chr(hex '0a'). #Type: SRATIONAL
        chr(hex '00').
        chr(hex '01'). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '4a'). #Offset Value
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_b_mpa_convergence_angle = $segref_left->create_record(
        'MPA_CONVERGENCE_ANGLE', 
        $BYTE,
        \$MPA_CONVERGENCE_ANGLE_B,
        12
        );

    #MP Baseline Length (5.2.4.7)
    my $MPA_BASELINE_LENGTH_B =
        chr(hex '06'). #Tag
        chr(hex 'b2').
        chr(hex '05'). #Type: RATIONAL
        chr(hex '00').
        chr(hex '01'). #Count
        chr(hex '00').
        chr(hex '00').
        chr(hex '00').
        chr(hex '52'). #offset value
        chr(hex '00').
        chr(hex '00').
        chr(hex '00');

    my $recref_b_mpa_baseline_length = $segref_left->create_record(
        'MPA_BASELINE_LENGTH', 
        $BYTE,
        \$MPA_BASELINE_LENGTH_B,
        12
        );

    ##########################################
    #update is not supported for APP2. 
    #So write the records manually in the buffer

    #write the records into the left segment buffer
    $buffer_left .= $recref_mp_endian->get();
    $buffer_left .= $recref_offset_to_first_ifd->get();

    #MPI are only for the first image
    $buffer_left .= $recref_mpi_count->get();
    $buffer_left .= $recref_mpi_version->get();
    $buffer_left .= $recref_mpi_number_of_images->get();
    $buffer_left .= $recref_mpe_tag->get();
    $buffer_left .= $recref_next_ifd_offset->get();

    for(my $i=0;$i<$NUMBER_OF_IMAGES;$i++){
        $buffer_left .= $recref_mpi_values[$i]->get();
    }

    $buffer_left .= $recref_mpa_count->get();
    $buffer_left .= $recref_mpa_individual_image_number->get();
    $buffer_left .= $recref_mpa_base_viewpoint_number->get();
    $buffer_left .= $recref_mpa_convergence_angle->get();
    $buffer_left .= $recref_mpa_baseline_length->get();
    $buffer_left .= $recref_next_ifd_offset_null->get();
    $buffer_left .= $recref_mpa_values->get();

    #write the records into the right segment buffer
    #same data as in the left buffer
    $buffer_right  = $recref_mp_fid->get();
    $buffer_right .= $recref_mp_endian->get();
    $buffer_right .= $recref_offset_to_first_ifd->get();

    $buffer_right .= $recref_b_mpa_count->get();
    $buffer_right .= $recref_mpi_version->get();  #MPI version = MPA version for second image
    $buffer_right .= $recref_b_mpa_individual_image_number->get();
    $buffer_right .= $recref_mpa_base_viewpoint_number->get();
    $buffer_right .= $recref_b_mpa_convergence_angle->get();
    $buffer_right .= $recref_b_mpa_baseline_length->get();
    $buffer_right .= $recref_next_ifd_offset_null->get();
    $buffer_right .= $recref_mpa_values->get();


    #save the files
    $file_left->save($filename_out);
    $file_right->save('tmp');
}
########################################
=pod
    concat left and right to left
=cut
sub concatfiles{
    my ($filename_left, $filename_right) = @_;
    
    open ( FH_LEFT, ">>", $filename_left ) 
        or die "Could not open file: $!";
	binmode (FH_LEFT);									#	herbw@wiskit.com

    open ( FH_RIGHT, "<", $filename_right ) 
        or die "Could not open file: $!";
	binmode (FH_RIGHT);									#	herbw@wiskit.com

    while ( my $line = <FH_RIGHT> ) {
        print FH_LEFT $line;
    }

    close (FH_LEFT);										#	herbw@wiskit.com
    close (FH_RIGHT);										#	herbw@wiskit.com
}

sub usage
{
    print "Convert jpg files into an MPO stereo image file
June 2015 Olivier Drevet. 
Released under the GNU GPL v3 license (see LICENSE file). 

usage: perl $0 -l left.jpg -r right.jpg -o out.MPO

Command line arguments: 
MANDATORY COMMAND LINE ARGUMENTS: 
-left  or -l FILENAME : left file
-right or -r FILENAME : right file

OPTIONAL COMMAND LINE ARGUMENTS: 
-out      or -o FILENAME : output file (default out.MPO)
-baseline or -b VALUE    : baseline in milimeters  (default 77)
-help     or -h          : Display this message and quit
";
}
