/*-----------------------------------------------------------------------------
 *	NAME
 *		lrCompare.cc - Compare Exif data to match stereoscopic pairs of photos
 *
 *		Copyright (C) 2016-2018 Herb Weiner <herbw@wiskit.com>
 *		Licensed under Creative Commons CC BY-SA 4.0.
 *		https://creativecommons.org/licenses/by-sa/4.0/legalcode
 *
 *	DESCRIPTION
 *		This software uses exiftool to compare and match corresponding
 *		left and right stereoscopic pairs of photos.
 *
 *	Building
 *		g++ -g -o lrCompare lrCompare.cc
 *		cl lrCompare.cc
 *
 *	CAVEATS:
 *		This software assumes that the file modification time matches the
 *		exif time on all files.
 *
 *		This software assumes that creation times monotonically increase with
 *		increasing sequence numbers.
 *-----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/timeb.h>  
#include <sys/types.h>
#include <sys/stat.h>
#include <cstdlib>

#if defined(_WIN32) || defined(_WIN64) 
	#include <direct.h>
	#define	sys_errlist	_sys_errlist
	#define	snprintf	_snprintf
	#define	vsnprintf	_vsnprintf
	#define	strcasecmp	_stricmp
	#define	strncasecmp	_strnicmp
	#define	getcwd		_getcwd
	#define	stat		_stat
	#define	popen		_popen
	#define	pclose		_pclose
	#define struct_stat	struct _stat64i32
	const char					OPTION = '/';
	const char					DIRECTORY = '\\';
	const char					PATHSEP = ';';
	const bool					WINDOWS = true;
#else
	#include <unistd.h>
	extern int					errno;
	extern const char * const	sys_errlist[];
	#define	struct_stat	struct stat
	const char					OPTION = '-';
	const char					DIRECTORY = '/';
	const char					PATHSEP = ':';
	const bool					WINDOWS = false;
#endif

static char						my_name [1024];
static char						opt_format [256];
static const char				*opt_left = "Raw-Left";
static const char				*opt_right = "Raw-Right";

/*-----------------------------------------------------------------------------
 *	NAME
 *		trim - Remove trailing blanks, tabs, and newlines
 *
 *	SYNOPSIS
 *		static void
 *		trim(
 *			char			*buffer)			- Buffer to be trimmed
 *
 *	RETURN VALUE
 *		None.
 *
 *	DESCRIPTION
 *		This function removes trailing whitespace from a null terminated
 *		string.
 *-----------------------------------------------------------------------------
 */

static void trim (char *buffer)
{
	int					i = strlen (buffer) - 1;

	while ((i >= 0)
	  && ((buffer[i] == ' ') || (buffer[i] == '\t') || (buffer[i] == '\n') || (buffer[i] == '\r') || (buffer[i] == '\f')))
		buffer[i--] = '\0';
}

/*-----------------------------------------------------------------------------
 *	NAME
 *		parseSequence - Parse a Sequence such as 102-137 or 105=106
 *
 *	SYNOPSIS
 *		static void
 *		parseSequence(
 *			char			*buffer,			- Buffer to be trimmed
 *			int				*n1,				- First Number
 *			int				*n2,				- Second Number
 *			char			*delim,				- Delimiter
 *			bool			wsOK)				- White Space OK
 *
 *	RETURN VALUE
 *		Returns integers n1 and n2 and the delimiter character (- or =).
 *		In the event of an error, n1 and/or n2 will be zero.
 *
 *	DESCRIPTION
 *		This function parses a parameter sequence of Start-End (e.g 102-137)
 *		or Left=Right (e.g. 105=106).
 *-----------------------------------------------------------------------------
 */

static void parseSequence (const char *cp, int *n1, int *n2, char *delim, bool wsOK)
{
	int							len;

	*n1 = 0;
	*n2 = 0;
	*delim = '?';
	len = strlen (cp);

	if (wsOK)
	{
		while ((*cp == ' ') || (*cp == '\t'))
			cp++;
	}

	if ((cp[0] != '-') && (cp[0] != '=') && (cp[len-1] != '-') && (cp[len-1] != '='))
	{
		while ((*cp >= '0') && (*cp <= '9'))
		{
			*n1 = *n1 * 10 + *cp - '0';
			cp++;
		}

		if (wsOK)
		{
			while ((*cp == ' ') || (*cp == '\t'))
				cp++;
		}


		if ((*cp == '-') || (*cp == '=') || (*cp == ','))
		{
			*delim = *cp++;

			if (wsOK)
			{
				while ((*cp == ' ') || (*cp == '\t'))
					cp++;
			}
		}

		while ((*cp >= '0') && (*cp <= '9'))
		{
			*n2 = *n2 * 10 + *cp - '0';
			cp++;
		}

		if (*cp != '\0')
			*n2 = 0;
	}
}

/*-----------------------------------------------------------------------------
 *	NAME
 *		create_sprintf_format - Create format string for sprintf
 *
 *	SYNOPSIS
 *		static void
 *		create_sprintf_format(
 *			char			*result,			- Buffer to be trimmed
 *			const char		*pattern)			- Filename Pattern
 *
 *	RETURN VALUE
 *		None.
 *
 *	DESCRIPTION
 *		This function creates a format string for sprintf.
 *-----------------------------------------------------------------------------
 */

void create_sprintf_format (char *result, const char *pattern)
{
	const char					*cp;
	const char					*ep;


	cp = pattern;
	while ((*cp != '\0') && (*cp != '#'))
		cp++;

	if (*cp == '\0')
	{
		*result = '\0';
		return;
	}

	ep = cp;

	while (*ep == '#')
		ep++;

	sprintf (result, "%%s/%.*s%%0%d.%dd%s", (int) (cp - pattern), pattern, (int) (ep - cp), (int) (ep - cp), ep);
}
/*-----------------------------------------------------------------------------
 *	NAME
 *		format_delta_t - Format the Delta Time between Two Files
 *
 *	SYNOPSIS
 *		static void
 *		format_delta_t(
 *			char			*buffer,			- Buffer to be trimmed
 *			int				left,				- Left File Number
 *			int				right,				- Right File Number
 *			time_t			timeLeft,			- Modification Time Left File
 *			time_t			timeRight,			- Modification Time Right File
 *			int				mode)				- 0 (L/R), 1 (L/L) or 2 (R/R)
 *
 *	RETURN VALUE
 *		None.
 *
 *	DESCRIPTION
 *		This function formats the Delta Time between two files.
 *-----------------------------------------------------------------------------
 */

void format_delta_t (char *buffer, int left, int right, time_t timeLeft, time_t timeRight, int mode = 0)
{
	struct tm					*my_time;
	char						time_left_string [256];
	char						time_right_string [256];
	char						left_fn [256];
	char						right_fn [256];

	switch (mode)
	{
		case 0:					//	Delta between Left and Right
			sprintf (left_fn, opt_format, opt_left, left);
			sprintf (right_fn, opt_format, opt_right, right);
			break;

		case 1:					//	Delta between Left and Left
			sprintf (left_fn, opt_format, opt_left, left);
			sprintf (right_fn, opt_format, opt_left, right);
			break;

		case 2:					//	Delta between Right and Right
			sprintf (left_fn, opt_format, opt_right, left);
			sprintf (right_fn, opt_format, opt_right, right);
			break;

		default:
			sprintf (left_fn, opt_format, opt_left, left);
			sprintf (right_fn, opt_format, opt_right, right);
			break;
	}

	my_time = localtime (&timeLeft);
	strftime (time_left_string, sizeof (time_left_string), "%a %Y-%m-%d %T", my_time);

	my_time = localtime (&timeRight);
	strftime (time_right_string, sizeof (time_right_string), "%a %Y-%m-%d %T", my_time);

	if (timeLeft > timeRight)
		sprintf (buffer, "Image %s (%s) is Newer than Image %s (%s) by %d seconds", left_fn, time_left_string, right_fn, time_right_string, (int) (timeLeft - timeRight));
	else if (timeRight > timeLeft)
		sprintf (buffer, "Image %s (%s) is Older than Image %s (%s) by %d seconds", left_fn, time_left_string, right_fn, time_right_string, (int) (timeRight - timeLeft));
	else
		sprintf (buffer, "Image %s (%s) and Image %s (%s) have Matching Creation Times", left_fn, time_left_string, right_fn, time_right_string);
}

/*-----------------------------------------------------------------------------
 *	NAME
 *		main - lrCompare main function
 *
 *	SYNOPSIS
 *		int
 *		main(
 *			int				argc,				- Number of Arguments
 *			const char		*argv[])			- Argument Vector
 *
 *	RETURN VALUE
 *		0 if No Errors.
 *		1 if Fatal Error.
 *
 *	DESCRIPTION
 *		This is the main function of lrCompare.
 *-----------------------------------------------------------------------------
 */

int main (int argc, const char *argv[])
{
	int							i;
	int							j;
	int							k;
	int							first;
	int							last;
	int							count;
	int							current;
	char						left_name [256];
	char						right_name [256];
	char						left_temp [256];
	char						right_temp [256];
	char						old_name [256];
	char						new_name [256];
	char						first_fn [256];
	char						last_fn [256];
	char						minimum [64];
	char						maximum [64];
	int							opt_help = 0;
	int							opt_debug = 0;
	int							opt_first = 0;
	int							opt_last = 0;
	int							opt_ignore = 0;
	int							opt_number = 0;
	int							opt_rename = 0;
	int							opt_prompt = 0;
	int							len;
	int							left_len;
	int							right_len;
	int							result;
	int							left_result;
	int							right_result;
	int							opt_verbose = 0;
	int							nErrors = 0;
	long						delta_t;
	long						expected_delta = 0;
	long						time_threshhold = 10;
	char						*cpl;
	char						*cpr;
	char						*epl;
	char						*epr;
	const char					*cp;
	const char					*ep;
	const char					*opt_time = "10";
	const char					*opt_pattern = "IMGP####.PEF";
	struct_stat					left_stat;
	struct_stat					right_stat;
	struct tm					*my_tm;
	char						cmd_left [1024];
	char						cmd_right [1024];
	char						exif_left [4096];
	char						exif_right [4096];
	char						delta_left [4096];
	char						delta_right [4096];
	char						left_time [128];
	char						right_time [128];
	char						buffer [1024];
	char						input [256];
	char						delim;
	const char					*exif_labels [] =
								{
									"ImageWidth",
									"ImageHeight",
									"FocalLength",
									"FocusRangeIndex",
									"ISO",
									"ExposureTime",
									"FNumber",
									"ExposureProgram",
									"PictureMode",
									"PictureMode2",
									"FocusMode",
									"MeteringMode",
									"FlashMode",
								};
	FILE						*p;
	const int					MAX_FILES = 4096;
	time_t						times_left [MAX_FILES] = {0};
	time_t						times_right [MAX_FILES] = {0};
	time_t						times_left_original [MAX_FILES] = {0};
	time_t						times_right_original [MAX_FILES] = {0};
	time_t						time_current_left = 0;
	time_t						time_current_right = 0;
	time_t						time_previous_left = 0;
	time_t						time_previous_right = 0;
	int							left_right_index [MAX_FILES] = {0};
	int							left_index_previous = 0;
	int							right_index_previous = 0;
	int							left_right_index_previous;
	int							left_right_index_current;
	int							delete_left = 0;
	int							delete_right = 0;
	int							rename_left = 0;
	int							rename_right = 0;
	int							pair_left = 0;
	int							pair_right = 0;
	int							best_index;
	int							best_delta;
	bool						found_match;

	typedef struct
	{
		const char * const	name;
		void				*value;
		const char * const	help;
	}
	option_structure;

	static option_structure option_list [] =
	{
		{ "-?",	&opt_help,				"Display these help messages"                       },
		{ "-h",	&opt_help,				"Display these help messages"                       },
		{ "-d",	&opt_debug,				"Debug Output"                                      },
		{ "-f",	&opt_first,				"Use First Pair as Time Reference"                  }, 
		{ "-i",	&opt_ignore,			"Ignore Errors"                                     },
		{ "=lf",&opt_left,				"Left Source Folder"                                },
		{ "-l",	&opt_last,				"Use Last Pair as Time Reference"                   },
		{ "-n",	&opt_number,			"Match Based on Image Number rather than Date/Time" },
		{ "=p",	&opt_pattern,			"Filename Pattern"                                  },
		{ "-pr",&opt_prompt,			"Prompt for First, Last, and Match"                 },
		{ "=rf",&opt_right,				"Right Source Folder"                               },
		{ "-r",	&opt_rename,			"Delete and Rename Files to Match Left and Right"   },
		{ "=t",	&opt_time,				"Time Difference Threshhold"                        },
		{ "-v",	&opt_verbose,			"Verbose Output"                                    },
	};
	int	number_of_options = sizeof (option_list) / sizeof (option_structure);

	/*-------------------------------------------------------------------------
	 *	Process Command Line and Options
	 *-------------------------------------------------------------------------
	 */

	cp = strrchr (argv[0], DIRECTORY);
	if (cp == (const char *) NULL)
		cp = argv[0];
	else
		cp++;

	ep = strrchr (cp, '.');
	if (ep == (const char *) NULL)
		ep = strchr (cp, '\0');

	strncpy (my_name, cp, ep - cp);
	my_name [ep - cp] = '\0';

	argv++;
	argc--;

	while ((argc > 0) && (**argv == OPTION))
	{
		for (i = 0; i < number_of_options; i++)
		{
			if (option_list[i].name[0] == '-')
			{
				if (strcmp (option_list [i].name + 1, *argv + 1) == 0)
				{
					*((int *) (option_list [i].value)) ^= 1;
					argv++;
					argc--;
					break;
				}
			}
			if (option_list[i].name[0] == '=')
			{
				if (strcmp (option_list [i].name + 1, *argv + 1) == 0)
				{
					if (argc > 1)
					{
						argv++;
						argc--;
						*((const char **) (option_list [i].value)) = *argv;
						argv++;
						argc--;
						break;
					}
					else
						fprintf (stderr, "Error: required value missing for %s\n", *argv);
				}
			}
		}

		/*---------------------------------------------------------------------
		 *	If the option was not found, try a case-insensitive search.
		 *---------------------------------------------------------------------
		 */

		if (i == number_of_options)
		{
			for (i = 0; i < number_of_options; i++)
			{
				if (option_list[i].name[0] == '-')
				{
					if (strcasecmp (option_list [i].name + 1, *argv + 1) == 0)
					{
						*((int *) (option_list [i].value)) ^= 1;
						argv++;
						argc--;
						break;
					}
				}
				if (option_list[i].name[0] == '=')
				{
					if (strcasecmp (option_list [i].name + 1, *argv + 1) == 0)
					{
						if (argc > 1)
						{
							argv++;
							argc--;
							*((const char **) (option_list [i].value)) = *argv;
							argv++;
							argc--;
							break;
						}
						else
							fprintf (stderr, "Error: required value missing for %s\n", *argv);
					}
				}
			}
		}

		if (i == number_of_options)
		{
			fprintf (stderr, "Error: unrecognized option %s\n", *argv);
			opt_help = 1;
			break;
		}
	}

	if ((argc < 1) && (! opt_help) && (! opt_prompt))
	{
		/*---------------------------------------------------------------------
		 *	If you double click on the Windows .exe, it will open a Command
		 *	window that will close when the program terminates. Provide
		 *	a brief synopsis of the program, and wait for the user to Press
		 *	ENTER before closing the window.
		 *---------------------------------------------------------------------
		 */

		fprintf (stderr, "lrCompare Copyright (c) 2018 Herb Weiner <herbw@wiskit.com>. All rights reserved.\n");
		fprintf (stderr, "Compare and match left and right images based upon exif data\n");
		fprintf (stderr, "    using exiftool (https://www.sno.phy.queensu.ca/~phil/exiftool/).\n");
		fprintf (stderr, "Use %s %ch for more information.\n", my_name, OPTION);

		if (WINDOWS)
		{
			fprintf (stderr, "Press ENTER to continue . . .\n");
			fgets (input, sizeof (input), stdin);
		}

		exit (0);
	}

	if (argc == 1)
	{
		parseSequence (argv [0], &first, &last, &delim, false);
		if ((first <= 0) || (last <= 0) || (delim != '-'))
		{
			fprintf (stderr, "%s: Invalid Start-End Range: %s\n", my_name, argv[0]);
			opt_help = 1;
		}
	}
	else if (argc == 2)
	{
		parseSequence (argv [0], &first, &last, &delim, false);

		if (delim == '-')
		{
			if ((first <= 0) || (last <= 0))
			{
				fprintf (stderr, "%s: Invalid Start-End Range: %s\n", my_name, argv[0]);
				opt_help = 1;
			}

			parseSequence (argv [1], &pair_left, &pair_right, &delim, false);

			if ((pair_left <= 0) || (pair_right <= 0) || (delim != '='))
			{
				fprintf (stderr, "%s: Invalid Left=Right Pairing: %s\n", my_name, argv[1]);
				opt_help = 1;
			}
		}
		else if (delim == '=')
		{
			pair_left = first;
			pair_right = last;

			if ((pair_left <= 0) || (pair_right <= 0))
			{
				fprintf (stderr, "%s: Invalid Left=Right Pairing: %s\n", my_name, argv[0]);
				opt_help = 1;
			}
			else
			{
				if (opt_first || opt_last || opt_number)
				{
					fprintf (stderr, "%s: Options -f and -l and -n are not valid with Left=Right Pairing: %s\n", my_name, argv[0]);
					opt_help = 1;
				}
			}

			parseSequence (argv [1], &first, &last, &delim, false);

			if ((first <= 0) || (last <= 0) || (delim != '-'))
			{
				fprintf (stderr, "%s: Invalid Start-End Range: %s\n", my_name, argv[1]);
				opt_help = 1;
			}
		}
		else
		{
			fprintf (stderr, "%s: Invalid Parameter: %s\n", my_name, argv[0]);
			opt_help = 1;
		}
	}
	else
	{
		if (opt_prompt)
		{
			while (1)
			{
				fprintf (stderr, "%s: Enter Starting and Ending Range, for example 1-10\n", my_name);
				cp = fgets (input, sizeof (input), stdin);
				if (cp != (const char *) NULL)
				{
					trim (input);

					parseSequence (input, &first, &last, &delim, true);

					if ((first > 0) && (last > 0))
						break;
				}

				fprintf (stderr, "%s: Entry was not valid\n", my_name);
			}

			if ((! opt_first) && (! opt_last) && (! opt_number))
			{
				while (1)
				{
					fprintf (stderr, "%s: Enter Numbers of Matching Left and Right Images, for example 3=5\n", my_name);
					cp = fgets (input, sizeof (input), stdin);
					if (cp != (const char *) NULL)
					{
						trim (input);

						parseSequence (input, &pair_left, &pair_right, &delim, true);

						if ((pair_left > 0) && (pair_right > 0))
							break;
					}

					fprintf (stderr, "%s: Entry was not valid\n", my_name);
				}
			}
		}
		else
		{
			if (!opt_help)
				fprintf (stderr, "%s: Start-End Range Must be Specified\n", my_name);

			opt_help = 1;
		}
	}

	if (opt_first + opt_last + opt_number > 1)
	{
		fprintf (stderr, "%s: Options -f and -l and -n may not be specified together\n", my_name);
		opt_help = 1;
	}

	if (opt_help)
	{
		fprintf (stderr, "%s: Compare Exif data to match stereoscopic pairs of photos\n", my_name);
		fprintf (stderr, "Copyright (C) 2016-2018 Herb Weiner <herbw@wiskit.com>\n");
		fprintf (stderr, "http://www.herbweiner.com/stereo-software.html\n");
		fprintf (stderr, "Licensed under Creative Commons CC BY-SA 4.0\n\n");

		fprintf (stderr, "usage: %s -options start-end [left=right]\n", my_name);
		fprintf (stderr, "options:\n");
		for (i = 0; i < number_of_options; i++)
		{
			if (option_list[i].name[0] == '-')
				fprintf (stderr, "  %c%s: %s [%s]\n",
							OPTION, option_list [i].name + 1, option_list [i].help,
							*((int *) (option_list [i].value)) ? "enabled" : "disabled");
			else
				fprintf (stderr, "  %c%s value: %s [%s]\n",
							OPTION, option_list [i].name + 1, option_list [i].help,
							*((const char **) (option_list [i].value)) ? (const char *) *((const char **) (option_list [i].value)) : "<UNDEFINED>");
		}

		exit (1);
	}

	if ((opt_time != (const char *) NULL) && (*opt_time != '\0'))
		time_threshhold = atoi (opt_time);

	if (last < first)
	{
		current = last;
		last = first;
		first = current;
	}

	count = last - first + 1;

	/*-------------------------------------------------------------------------
	 *	Create Pattern for Filenames
	 *-------------------------------------------------------------------------
	 */

	create_sprintf_format (opt_format, opt_pattern);

	/*-------------------------------------------------------------------------
	 *	Get Date and Time for all Left and Right Images
	 *-------------------------------------------------------------------------
	 */

	for (current = first; current <= last; current += 1)
	{
		sprintf (left_name, opt_format, opt_left, current);
		sprintf (right_name, opt_format, opt_right, current);

		left_result = stat (left_name, &left_stat);
		right_result = stat (right_name, &right_stat);

		if (left_result != -1)
		{
			times_left [current - first] = left_stat.st_mtime;
			times_left_original [current - first] = left_stat.st_mtime;

			if (left_stat.st_mtime < time_previous_left)
			{
				format_delta_t (buffer, current, left_index_previous, left_stat.st_mtime, time_previous_left, 1);
				fprintf (stderr, "%s: ERROR: %s but MUST be NEWER\n", my_name, buffer);
				nErrors++;
			}

			left_index_previous = current;
			time_previous_left = left_stat.st_mtime;
		}
		else
		{
			times_left [current - first] = 0;
			times_left_original [current - first] = 0;
		}

		if (right_result != -1)
		{
			times_right [current - first] = right_stat.st_mtime;
			times_right_original [current - first] = right_stat.st_mtime;

			if (right_stat.st_mtime < time_previous_right)
			{
				format_delta_t (buffer, current, right_index_previous, right_stat.st_mtime, time_previous_right, 2);
				fprintf (stderr, "%s: ERROR: %s but MUST be NEWER\n", my_name, buffer);
				nErrors++;
			}

			right_index_previous = current;
			time_previous_right = right_stat.st_mtime;
		}
		else
		{
			times_right [current - first] = 0;
			times_right [current - first] = 0;
		}
	}

	if (opt_first || opt_last || ((pair_left != 0) && (pair_right != 0)))
	{
		if (opt_first)
		{
			expected_delta = times_left [0] - times_right [0];
			format_delta_t (buffer, first, first, times_left [0], times_right [0]);	
		}
		else if (opt_last)
		{
			expected_delta = times_left [count] - times_right [count];
			format_delta_t (buffer, last, last, times_left [count], times_right [count]);	
		}
		else
		{
			expected_delta = times_left [pair_left - first] - times_right [pair_right - first];
			format_delta_t (buffer, pair_left, pair_right, times_left [pair_left - first], times_right [pair_right - first]);	
		}
	}
	else
		strcpy (buffer, "Will match by filename");

	if (opt_verbose || opt_prompt)
	{
		if (nErrors > 0)
		{
			if (opt_prompt)
			{
				fprintf (stderr, "%s: %d serious error(s) found. Continue at your own risk\n", my_name, nErrors);

				while (1)
				{
					fprintf (stderr, "%s: Enter 1 or 2:\n", my_name);
					fprintf (stderr, "    1: Ignore errors and Continue\n");
					fprintf (stderr, "    2: Quit\n");

					cp = fgets (input, sizeof (input), stdin);
					if (cp != (const char *) NULL)
					{
						trim (input);
						while ((*cp == ' ') || (*cp == '\t'))
							cp++;

						if ((cp[0] == '1') && (cp[1] == '\0'))
							break;
						else if ((cp[0] == '2') && (cp[1] == '\0'))
							exit (0);
					}

					fprintf (stderr, "%s: Entry was not valid\n", my_name);
				}
			}
			else
			{
				if (! opt_ignore)
				{
					fprintf (stderr, "%s: Quitting due to %d serious error(s) found. To ignore errors, specify %ci\n", my_name, nErrors, OPTION);
					exit (1);
				}
			}
		}

		sprintf (first_fn, opt_format + 3, first);
		sprintf (last_fn, opt_format + 3, last);
		fprintf (stderr, "%s: Will examine files from %s to %s\n", my_name, first_fn, last_fn);

		fprintf (stderr, "%s: %s\n", my_name, buffer);

		if ((expected_delta - time_threshhold) < 0)
			sprintf (minimum, "left file older than right file by %ld seconds", -(expected_delta - time_threshhold));
		else if ((expected_delta - time_threshhold) > 0)
			sprintf (minimum, "left file newer than right file by %ld seconds", (expected_delta - time_threshhold));
		else
			strcpy (minimum, "left file same time as right file");
 
		if ((expected_delta + time_threshhold) < 0)
			sprintf (maximum, "left file older than right file by %ld seconds", -(expected_delta + time_threshhold));
		else if ((expected_delta + time_threshhold) > 0)
			sprintf (maximum, "left file newer than right file by %ld seconds", (expected_delta + time_threshhold));
		else
			strcpy (maximum, "left file same time as right file");

		fprintf (stderr, "%s: Expected Delta: %ld, Time Threshhold: %ld (%s TO %s)\n", my_name, expected_delta, time_threshhold, minimum, maximum);

		if (opt_prompt)
		{
			while (1)
			{
				fprintf (stderr, "%s: Enter 1, 2, or 3:\n", my_name);
				fprintf (stderr, "    1: Delete and Rename files as necessary to make contents of %s and %s match (between %s and %s)\n", opt_left, opt_right, first_fn, last_fn);
				fprintf (stderr, "    2: Report differences between the contents of %s and %s (between %s and %s) but make NO changes\n", opt_left, opt_right, first_fn, last_fn);
				fprintf (stderr, "    3: Quit\n");

				cp = fgets (input, sizeof (input), stdin);
				if (cp != (const char *) NULL)
				{
					trim (input);
					while ((*cp == ' ') || (*cp == '\t'))
						cp++;

					if ((cp[0] == '1') && (cp[1] == '\0'))
					{
						opt_rename = 1;
						break;
					}
					else if ((cp[0] == '2') && (cp[1] == '\0'))
					{
						opt_rename = 0;
						break;
					}
					else if ((cp[0] == '3') && (cp[1] == '\0'))
						exit (0);
				}

				fprintf (stderr, "%s: Entry was not valid\n", my_name);
			}
		}
		else
		{
			if (opt_rename)
				fprintf (stderr, "%s: Will Delete and Rename files as necessary to make contents of %s and %s match (between %s and %s)\n", my_name, opt_left, opt_right, first_fn, last_fn);
			else
				fprintf (stderr, "%s: Will NOT alter the contents of %s and %s (between %s and %s)\n", my_name, opt_left, opt_right, first_fn, last_fn);
		}
	}

	left_right_index_previous = first - 1;

	/*-------------------------------------------------------------------------
	 *	Match Corresponding Left and Right Images by Date and Time or
	 *	Sequence Number.
	 *-------------------------------------------------------------------------
	 */

	for (current = first; current <= last; current++)
	{
		if (times_left [current - first] == 0)
			continue;

		for (i = left_right_index_previous + 1; i <= last; i++)
		{
			if (opt_number)
				found_match = (i == current);
			else
				found_match = (std::abs (times_left [current - first] - times_right [i - first] - expected_delta) <= time_threshhold);

			if (found_match)
			{
				/*-------------------------------------------------------------
				 *	Skip / Remove any Right Files in this Iteration.
				 *-------------------------------------------------------------
				 */

				for (j = left_right_index_previous + 1; j < i; j++)
				{
					time_current_right = times_right [j - first];

					if (time_current_right != 0)
					{
						my_tm = localtime (&(time_current_right));
						strftime (right_time, sizeof (right_time), "%a %d-%m-%Y %T", my_tm);

						best_delta = 86400;		//	One Day in Seconds
						best_index = 0;
						for (k = first; k <= last; k++)	//	This is not the most efficient possible search
						{
							delta_t = times_left_original [k - first] - times_right_original [j - first] - expected_delta;
							if (std::abs (delta_t) < best_delta)
							{
								best_delta = std::abs (delta_t);
								best_index = k;
							}
						}
						time_current_left = times_left_original [best_index - first];
						my_tm = localtime (&(time_current_left));
						strftime (left_time, sizeof (left_time), "%a %d-%m-%Y %T", my_tm);

						sprintf (left_temp, opt_format, opt_left, best_index);
						sprintf (right_name, opt_format, opt_right, j);

						fprintf (stderr, "%s: Skipping %s (%s), Closest Match is %s (%s), Differ by %d seconds (adjusted)\n", my_name, right_name, right_time, left_temp, left_time, best_delta);

						if (opt_rename)
						{
							fprintf (stderr, "%s: Delete %s\n", my_name, right_name);
							result = unlink (right_name);
							if (result == -1)
								fprintf (stderr, "%s: unlink (%s) failed <%s>\n", my_name, right_name, sys_errlist [errno]);

							delete_right++;
						}

						times_right [j - first] = 0;
					}
				}

				time_current_left = times_left [current - first];
				time_current_right = times_right [i - first];

				my_tm = localtime (&(time_current_left));
				strftime (left_time, sizeof (left_time), "%a %d-%m-%Y %T", my_tm);

				my_tm = localtime (&(time_current_right));
				strftime (right_time, sizeof (right_time), "%a %d-%m-%Y %T", my_tm);

				if (opt_verbose || (i != current))
				{
					delta_t = times_left [current - first] - times_right [i - first] - expected_delta;
					sprintf (left_temp, opt_format, opt_left, current);
					sprintf (right_temp, opt_format, opt_right, i);
					fprintf (stderr, "%s: Matching %s (%s) to %s (%s), Differ by %ld seconds\n", my_name, left_temp, left_time, right_temp, right_time, std::abs (delta_t));
				}

				left_right_index [current - first] = i;

				left_right_index_previous = i;

				if (i != current)
				{
					if (opt_rename)
					{
						if (i < current)
						{
							/*-------------------------------------------------
							 *	The Right Image (i) has a Lower Number than the
							 *	Left Image (current). Rename the Left (Higher
							 *	Number) Image to Match the Number on the Right
							 *	(Lower Number) Image.
							 *-------------------------------------------------
							 */

							if (left_right_index [i - first] != 0)
								fprintf (stderr, "%s: ERROR: left_right_index [%d] is %d\n", my_name, i, left_right_index [i - first]);
							else
								left_right_index [i - first] = i;

							left_right_index [current - first] = 0;
							//	No change required to left_right_index_previous.

							sprintf (old_name, opt_format, opt_left, current);
							sprintf (new_name, opt_format, opt_left, i);

							times_left [i - first] = times_left [current - first];
							times_left [current - first] = 0;

							rename_left++;
						}
						else
						{
							/*-------------------------------------------------
							 *	The Right Image (i) has a higher Number than
							 *	the Left Image (current). Rename the Right
							 *	(Higher Number) Image to Match the Number
							 *	on the Left (Lower Number) Image.
							 *-------------------------------------------------
							 */

							sprintf (old_name, opt_format, opt_right, i);
							sprintf (new_name, opt_format, opt_right, current);

							left_right_index [current - first] = current;
							left_right_index_previous = current;

							times_right [current - first] = times_right [i - first];
							times_right [i - first] = 0;

							if (opt_debug)
								fprintf (stderr, "%s: DEBUG: about to rename %s to %s. Setting times_right [%d] = 0. Setting times_right [%d] = %ld\n", my_name, old_name, new_name, i, current, (long) times_right [current - first]);

							rename_right++;
						}

						fprintf (stderr, "%s: Rename %s to %s\n", my_name, old_name, new_name);
						result = rename (old_name, new_name);
						if (result == -1)
							fprintf (stderr, "%s: rename (%s, %s) failed <%s>\n", my_name, old_name, new_name, sys_errlist [errno]);
					}
				}

				break;
			}
		}

		/*---------------------------------------------------------------------
		 *	Skip / Remove any Left Files in this Iteration.
		 *---------------------------------------------------------------------
		 */

		if (left_right_index [current - first] == 0)
		{
			time_current_left = times_left [current - first];

			if (time_current_left != 0)
			{
				my_tm = localtime (&(time_current_left));
				strftime (left_time, sizeof (left_time), "%a %d-%m-%Y %T", my_tm);

				best_delta = 86400;		//	One Day in Seconds
				best_index = 0;
				for (k = first; k <= last; k++)		//	This is not the most efficient possible search
				{
					delta_t = times_left_original [current - first] - times_right_original [k - first] - expected_delta;
					if (std::abs (delta_t) < best_delta)
					{
						best_delta = std::abs (delta_t);
						best_index = k;
					}
				}
				time_current_right = times_right_original [best_index - first];
				my_tm = localtime (&(time_current_right));
				strftime (right_time, sizeof (right_time), "%a %d-%m-%Y %T", my_tm);

				sprintf (left_name, opt_format, opt_left, current);
				sprintf (right_temp, opt_format, opt_right, best_index);

				fprintf (stderr, "%s: Skipping %s (%s), Closest Match is %s (%s), Differ by %d seconds (adjusted)\n", my_name, left_name, left_time, right_temp, right_time, best_delta);

				if (opt_rename)
				{
					fprintf (stderr, "%s: Delete %s\n", my_name, left_name);
					result = unlink (left_name);
					if (result == -1)
						fprintf (stderr, "%s: unlink (%s) failed <%s>\n", my_name, left_name, sys_errlist [errno]);

					delete_left++;
				}

				times_left [current - first] = 0;
			}
		}
	}

	/*-------------------------------------------------------------------------
	 *	Skip / Remove any Remaining Right Files after the Final Match.
	 *-------------------------------------------------------------------------
	 */

	for (j = left_right_index_previous + 1; j <= last; j++)
	{
		time_current_right = times_right [j - first];

		if (time_current_right != 0)
		{
			my_tm = localtime (&(time_current_right));
			strftime (right_time, sizeof (right_time), "%a %d-%m-%Y %T", my_tm);

			best_delta = 86400;		//	One Day in Seconds
			best_index = 0;
			for (k = first; k <= last; k++)	//	This is not the most efficient possible search
			{
				delta_t = times_left_original [k - first] - times_right_original [j - first] - expected_delta;
				if (std::abs (delta_t) < best_delta)
				{
					best_delta = std::abs (delta_t);
					best_index = k;
				}
			}
			time_current_left = times_left_original [best_index - first];
			my_tm = localtime (&(time_current_left));
			strftime (left_time, sizeof (left_time), "%a %d-%m-%Y %T", my_tm);

			sprintf (left_temp, opt_format, opt_left, best_index);
			sprintf (right_name, opt_format, opt_right, j);

			fprintf (stderr, "%s: Skipping %s (%s), Closest Match is %s (%s), Differ by %d seconds (adjusted)\n", my_name, right_name, right_time, left_temp, left_time, best_delta);

			if (opt_rename)
			{
				fprintf (stderr, "%s: Delete %s\n", my_name, right_name);
				result = unlink (right_name);
				if (result == -1)
					fprintf (stderr, "%s: unlink (%s) failed <%s>\n", my_name, right_name, sys_errlist [errno]);

				delete_right++;
			}

			times_right [j - first] = 0;
		}
	}

	/*-------------------------------------------------------------------------
	 *	Report Statistics on Files Renamed and Deleted
	 *-------------------------------------------------------------------------
	 */

	if (opt_rename)
	{
		if (rename_left + rename_right + delete_left + delete_right == 0)
			fprintf (stderr, "%s: No files renamed or deleted\n", my_name);
		else
			fprintf (stderr, "%s: Renamed %d Left and %d Right Files. Deleted %d Left and %d Right Files.\n", my_name, rename_left, rename_right, delete_left, delete_right);
	}

	/*-------------------------------------------------------------------------
	 *	Report Corresponding Left and Right Images for Debugging Purposes
	 *-------------------------------------------------------------------------
	 */

	if (opt_debug)
	{
		for (current = first; current <= last; current += 1)
		{
			left_right_index_current = left_right_index [current - first];
			time_current_left = times_left [current - first];

			if (left_right_index_current == 0)
				time_current_right = times_right [current - first];
			else
				time_current_right = times_right [left_right_index_current - first];

			if (time_current_left == 0)
			{
				fprintf (stderr, "%s: DEBUG: Left: %d, Right: %d, Left Time = 0, Right Time = %ld\n", my_name, current, left_right_index_current, (long) time_current_right);
				continue;
			}

			if (time_current_right == 0)
				fprintf (stderr, "%s: Left: %d, Right: %d, Right Time = 0\n", my_name, current, left_right_index_current);
			else
				fprintf (stderr, "%s: Left: %d, Right: %d\n", my_name, current, left_right_index_current);
		}
	}

	/*-------------------------------------------------------------------------
	 *	Process Pairs of Left and Right Images Comparing EXIF Data
	 *-------------------------------------------------------------------------
	 */

	for (current = first; current <= last; current++)
	{
		left_right_index_current = left_right_index [current - first];
		time_current_left = times_left [current - first];
		time_current_right = times_right [left_right_index_current - first];

		if ((time_current_left == 0) || (time_current_right == 0))
			continue;

		my_tm = localtime (&(time_current_left));
		strftime (left_time, sizeof (left_time), "%a %d-%m-%Y %T", my_tm);

		my_tm = localtime (&(time_current_right));
		strftime (right_time, sizeof (right_time), "%a %d-%m-%Y %T", my_tm);

		sprintf (left_name, opt_format, opt_left, current);
		sprintf (right_name, opt_format, opt_right, left_right_index_current);

		delta_t = std::abs ((time_current_left  - time_current_right) - expected_delta);

		sprintf (cmd_left, "exiftool -T -ImageWidth -ImageHeight -LensFocalLength -FocusRangeIndex -ISO -ExposureTime -FNumber -ExposureProgram -PictureMode -PictureMode2 -FocusMode -MeteringMode -FlashMode %s", left_name);
		sprintf (cmd_right, "exiftool -T -ImageWidth -ImageHeight -LensFocalLength -FocusRangeIndex -ISO -ExposureTime -FNumber -ExposureProgram -PictureMode -PictureMode2 -FocusMode -MeteringMode -FlashMode %s", right_name);

		*exif_left = '\0';
		*exif_right = '\0';

		p = popen (cmd_left, "r");
		if (p == (FILE *) NULL)
			fprintf (stderr, "%s: popen (exif left) failed <%s>", my_name, sys_errlist [errno]);
		else
		{
			fgets (exif_left, sizeof (exif_left), p);
			trim (exif_left);
			pclose (p);
		}

		p = popen (cmd_right, "r");
		if (p == (FILE *) NULL)
			fprintf (stderr, "%s: popen (exif right) failed <%s>", my_name, sys_errlist [errno]);
		else
		{
			fgets (exif_right, sizeof (exif_right), p);
			trim (exif_right);
			pclose (p);
		}

		if ((delta_t > time_threshhold) && (! opt_number))
		{
			fprintf (stderr, "%s: Mismatch: %s (%s) and %s (%s) differ by %ld seconds\n", my_name, left_name, left_time, right_name, right_time, delta_t);
		}
		else
		{
			if (opt_verbose)
				fprintf (stderr, "%s: %s (%s) and %s (%s) Match (differ by %ld seconds)\n", my_name, left_name, left_time, right_name, right_time, delta_t);

			/*-----------------------------------------------------------------
			 *	If not opt_verbose, go tab by tab, compare, and only display
			 *	fields that differ
			 *-----------------------------------------------------------------
			 */

			if (opt_verbose || (strcmp (exif_left, exif_right) != 0))
			{
				*delta_left = '\0';
				*delta_right = '\0';
				epl = exif_left;
				epr = exif_right;
				i = 0;

				while (1)
				{
					cpl = epl;
					cpr = epr;

					if ((cpl == (const char *) NULL) || (cpr == (const char *) NULL))
						break;
			
					epl = strchr (cpl, '\t');
					if (epl != (const char *) NULL)
						*epl++ = '\0';

					epr = strchr (cpr, '\t');
					if (epr != (const char *) NULL)
						*epr++ = '\0';

					if (strcmp (cpl, cpr) != 0)
					{
						strcat (delta_left, ", ");
						strcat (delta_right, ", ");

						left_len = strlen (delta_left);
						right_len = strlen (delta_right);					

						if (left_len > right_len)
							sprintf (delta_right + right_len, "%*s", left_len - right_len, " ");
						else if (right_len > left_len)
							sprintf (delta_left + left_len, "%*s", right_len - left_len, " ");

						sprintf (buffer, "%s: %s", exif_labels[i], cpl);
						strcat (delta_left, buffer);

						sprintf (buffer, "%s: %s", exif_labels[i], cpr);
						strcat (delta_right, buffer);
					}

					i++;
				}

				fprintf (stderr, "%s: %s  %s%s\n", my_name, left_time, left_name, delta_left);
				fprintf (stderr, "%s: %s %s%s\n", my_name, right_time, right_name, delta_right);
			}
		}
	}

	return (0);
}
