/*-----------------------------------------------------------------------------
 *	NAME
 *		raw2jpg - Convert raw to jpg file (using ImageMagick)
 *
 *		Copyright (C) 2016-2018 Herb Weiner <herbw@wiskit.com>
 *		Licensed under Creative Commons CC BY-SA 4.0.
 *		https://creativecommons.org/licenses/by-sa/4.0/legalcode
 *
 *	DESCRIPTION
 *		This software uses the ImageMagick convert command to convert raw
 *		files to jpg.
 *
 *	Building
 *		g++ -g -o raw2jpg raw2jpg.cc
 *		cl raw2jpg.cc
 *
 *	Usage
 *		raw2jpg /d destination file...
 *-----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>  
#include <sys/timeb.h>  
#include <sys/types.h>
#include <sys/stat.h>

#if defined(_WIN32) || defined(_WIN64) 
	#include <direct.h>
	#define	sys_errlist	_sys_errlist
	#define	snprintf	_snprintf
	#define	vsnprintf	_vsnprintf
	#define	strcasecmp	_stricmp
	#define	strncasecmp	_strnicmp
	#define	getcwd		_getcwd
	#define	stat		_stat
	#define struct_stat	struct _stat64i32
	const char					OPTION = '/';
	const char					DIRECTORY = '\\';
	const char					PATHSEP = ';';
	const char					*opt_mpo_command = "perl mpo.pl";
	const bool					WINDOWS = true;
#else
	#include <unistd.h>
	extern int					errno;
	extern const char * const	sys_errlist[];
	#define	struct_stat	struct stat
	const char					OPTION = '-';
	const char					DIRECTORY = '/';
	const char					PATHSEP = ':';
	const char					*opt_mpo_command = "mpo.pl";
	const bool					WINDOWS = false;
#endif

static char						my_name [1024];

static int						opt_verbose = 0;
static int						opt_concise = 0;
static int						opt_rotate_psd = 0;
static const char				*opt_jpg = "JPG";
static const char				*opt_left = "Raw-Left";
static const char				*opt_right = "Raw-Right";
static const char				*opt_left_rotation = "";
static const char				*opt_right_rotation = "";
static const char				*opt_rotation = "";
static const char				*opt_suffix = "";
static const char				*opt_mpo = "";

/*-----------------------------------------------------------------------------
 *	NAME
 *		trim - Remove trailing blanks, tabs, and newlines
 *
 *	SYNOPSIS
 *		static void
 *		trim(
 *			char			*buffer)			- Buffer to be trimmed
 *
 *	RETURN VALUE
 *		None.
 *
 *	DESCRIPTION
 *		This function removes trailing whitespace from a null terminated
 *		string.
 *-----------------------------------------------------------------------------
 */

static void trim (char *buffer)
{
	int							i = strlen (buffer) - 1;

	while ((i >= 0)
	  && ((buffer[i] == ' ') || (buffer[i] == '\t') || (buffer[i] == '\n') || (buffer[i] == '\r') || (buffer[i] == '\f')))
		buffer[i--] = '\0';
}

/*-----------------------------------------------------------------------------
 *	NAME
 *		parse_fn - Split filename path into dirname, basename, and extension
 *
 *	SYNOPSIS
 *		void
 *		parse_fn(
 *			const char		*path,				- Complete Path (input)
 *			char			*dirname,			- Directory Component (output)
 *			char			*basename,			- Basename (output)
 *			char			*extension)			- Extension (output)
 *
 *	RETURN VALUE
 *		None.
 *
 *	DESCRIPTION
 *		This function Splits a filename path into dirname, basename, and
 *		extension components.
 *-----------------------------------------------------------------------------
 */

void parse_fn (const char *path, char *dirname, char *basename, char *extension)
{
	int							i;
	const char					*ep;
	const char					*bp;

	ep = strrchr (path, '.');
	if (ep == (const char *) NULL)
	{
		*extension = '\0';
		ep = strchr (path, '\0');
	}
	else
		strcpy (extension, ep + 1);

	bp = strrchr (path, DIRECTORY);
	if (bp == (const char *) NULL)
	{
		*dirname = '\0';

		strncpy (basename, path, ep - path);
		basename [ep - path] = '\0';
	}
	else
	{
		strncpy (dirname, path, bp - path);
		dirname [bp - path] = '\0';

		strncpy (basename, bp + 1, ep - bp - 1);
		basename [ep - bp - 1] = '\0';
	}
}

/*-----------------------------------------------------------------------------
 *	NAME
 *		check_dependency - Check to see whether target is older than dependency
 *
 *	SYNOPSIS
 *		int
 *		check_dependency(
 *			const char		*description,		- Description for Logging
 *			const char		*target,			- Complete Path (input)
 *			const char		*dependency)		- Complete Path (input)
 *
 *	RETURN VALUE
 *		-1: Error: Dependency does not exist.
 *		 0: Target is up to date.
 *		+1: Target is older than dependency and needs to be updated.
 *
 *	DESCRIPTION
 *		This function checks to see whether the target is older than the
 *		dependency, and thus must be updated.
 *
 *		This function is inspired by make. The result is:
 *		*	If dependency does not exist, return -1
 *		*	If dependency exists and target does not, +1
 *		*	If both exist and dependency is newer, +1
 *		*	Otherwise, return 0
 *
 *		If opt_verbose is set, log the result.
 *-----------------------------------------------------------------------------
 */

int check_dependency (const char *description, const char *target, const char *dependency)
{
	int							result;
	struct_stat					my_stat;
	long long					dependency_time;
	long long					target_time;
	time_t						my_time_t;
	struct tm					*my_time;
	char						dependency_time_string [256];
	char						target_time_string [256];

	result = stat (dependency, &my_stat);

	if (result == -1)
	{
		if (opt_verbose)
			fprintf (stderr, "%s: %s FAILED: %s: %s\n", my_name, description, dependency, sys_errlist [errno]);

		return (-1);
	}

	dependency_time = my_stat.st_mtime;

	/*-------------------------------------------------------------------------
	 *	If we have gotten to this point, the dependency exists
	 *-------------------------------------------------------------------------
	 */

	result = stat (target, &my_stat);

	if (result == -1)
	{
		if (opt_verbose)
			fprintf (stderr, "%s: %s UPDATE: %s: %s\n", my_name, description, target, sys_errlist [errno]);

		return (1);
	}

	target_time = my_stat.st_mtime;

	/*-------------------------------------------------------------------------
	 *	If we have gotten to this point, both the dependency and target exist
	 *-------------------------------------------------------------------------
	 */

	if (opt_verbose)
	{
		my_time_t = (time_t) dependency_time;
		my_time = localtime (&my_time_t);
		strftime (dependency_time_string, sizeof (dependency_time_string), "%a %Y-%m-%d %T", my_time);

		my_time_t = (time_t) target_time;
		my_time = localtime (&my_time_t);
		strftime (target_time_string, sizeof (target_time_string), "%a %Y-%m-%d %T", my_time);
	}

	if (dependency_time > target_time)
	{
		if (opt_verbose)
			fprintf (stderr, "%s: %s UPDATE: %s (%s) is newer than %s (%s)\n", my_name, description, dependency, dependency_time_string, target, target_time_string);

		return (1);
	}

	if (opt_verbose)
		fprintf (stderr, "%s: %s OK: %s (%s) is older than %s (%s)\n", my_name, description, dependency, dependency_time_string, target, target_time_string);

	return (0);
}

/*-----------------------------------------------------------------------------
 *	NAME
 *		process_file - Process one file
 *
 *	SYNOPSIS
 *		void
 *		process_file(
 *			const char		*path)				- File to Process
 *
 *	RETURN VALUE
 *		None.
 *
 *	DESCRIPTION
 *		This function processes one file.
 *
 *		The following logic is used:
 *		*	if check_dependency (RAW, PSD) == 1
 *		*		source = PSD
 *		*	else
 *		*		source = raw
 *		*	if check_dependency (JPG, source) == 1
 *		*		convert source to JPG
 *		*	if left does not exist or right does not exist
 *		*		return
 *		*	 if check_dependency (MPO, JPG-L) || check_dependency (MPO, JPG-R)
 *		*		create MPO from JPG-L and JPG-R
 *-----------------------------------------------------------------------------
 */

void process_file (const char *path)
{
	char						dirname [4096];
	char						basename [1024];
	char						extension [256];
	char						command [1024];
	char						time_string [256];
	int							result;
	int							left_result;
	int							right_result;
	struct_stat					my_stat;
	struct tm					*my_time;
	const char					*source_path;
	const char					*suffix = opt_suffix;
	const char					*rotation = opt_rotation;
	char						raw_path [4096];
	char						psd_path [4096];
	char						jpg_path [4096];
	char						jpg_left [4096];
	char						jpg_right [4096];
	char						mpo_path [4096];

	parse_fn (path, dirname, basename, extension);

	sprintf (psd_path, "%s%c%s.%s", dirname, DIRECTORY, basename, "psd");

	if (strcasecmp (dirname, opt_left) == 0)
	{
		suffix = "-L";
		rotation = opt_left_rotation;
	}
	else if (strcasecmp (dirname, opt_right) == 0)
	{
		suffix = "-R";
		rotation = opt_right_rotation;
	}

	sprintf (jpg_left, "%s%c%s%s.%s", opt_jpg, DIRECTORY, basename, "-L", "jpg");
	sprintf (jpg_right, "%s%c%s%s.%s", opt_jpg, DIRECTORY, basename, "-R", "jpg");
	sprintf (jpg_path, "%s%c%s%s.%s", opt_jpg, DIRECTORY, basename, suffix, "jpg");

	result = check_dependency ("psd", path, psd_path);
	if (result == 1)
		source_path = psd_path;
	else
		source_path = path;

	result = check_dependency ("jpg", jpg_path, source_path);
	if (result == 1)
	{
		if ((rotation != (const char *) NULL) && (*rotation != '\0'))
			sprintf (command, "convert -rotate %s %s %s", rotation, source_path, jpg_path);
		else
			sprintf (command, "convert %s %s", source_path, jpg_path);

		if (opt_concise)
			fprintf (stderr, "%s --> %s\n", source_path, jpg_path);
		else
			fprintf (stderr, "\t%s\n", command);

		result = system (command);
	//	fprintf (stderr, "\tresult = %d\n", result);
	}

	/*-------------------------------------------------------------------------
	 *	If we have converted the specified raw or psd file to jpg. This might
	 *	be the left or right image of a stereo pair, or a non-stereoscopic
	 *	image.
	 *
	 *	If this is not part of a stereoscopic pair, we are done.
	 *-------------------------------------------------------------------------
	 */

	if (*suffix == '\0')
		return;

	/*-------------------------------------------------------------------------
	 *	If we get to this point, the image is part of a stereoscopic pair.
	 *	Attempt to convert the other half of the pair.
	 *-------------------------------------------------------------------------
	 */

	if (suffix [1] == 'L')
	{
		sprintf (raw_path, "%s%c%s.%s", opt_right, DIRECTORY, basename, extension);
		sprintf (psd_path, "%s%c%s.%s", opt_right, DIRECTORY, basename, "psd");

		sprintf (jpg_path, "%s%c%s%s.%s", opt_jpg, DIRECTORY, basename, "-R", "jpg");

		rotation = opt_right_rotation;
	}
	else
	{
		sprintf (raw_path, "%s%c%s.%s", opt_left, DIRECTORY, basename, extension);
		sprintf (psd_path, "%s%c%s.%s", opt_left, DIRECTORY, basename, "psd");

		sprintf (jpg_path, "%s%c%s%s.%s", opt_jpg, DIRECTORY, basename, "-L", "jpg");

		rotation = opt_left_rotation;
	}

	result = check_dependency ("psd (stereo)", raw_path, psd_path);
	if (result == 1)
		source_path = psd_path;
	else
		source_path = raw_path;

	result = check_dependency ("jpg (stereo)", jpg_path, source_path);
	if (result == 1)
	{
		if ((rotation != (const char *) NULL) && (*rotation != '\0'))
			sprintf (command, "convert -rotate %s %s %s", rotation, source_path, jpg_path);
		else
			sprintf (command, "convert %s %s", source_path, jpg_path);

		if (opt_concise)
			fprintf (stderr, "%s --> %s\n", source_path, jpg_path);
		else
			fprintf (stderr, "\t%s\n", command);

		result = system (command);
	//	fprintf (stderr, "\tresult = %d\n", result);
	}

	/*-------------------------------------------------------------------------
	 *	If we get to this point, we have converted both halves of a
	 *	stereowscopic pair. Attempt to create an MPO file.
	 *
	 *	CAVEAT: Is it possible that we have not done the conversion because
	 *	one of the dependent files was missing.
	 *-------------------------------------------------------------------------
	 */

	if ((opt_mpo == (const char *) NULL) || (*opt_mpo == '\0'))
		return;

	if ((opt_mpo_command == (const char *) NULL) || (*opt_mpo_command == '\0'))
		return;

	sprintf (mpo_path, "%s%c%s.%s", opt_mpo, DIRECTORY, basename, "mpo");

	left_result = check_dependency ("mpo left", mpo_path, jpg_left);
	right_result = check_dependency ("mpo right", mpo_path, jpg_right);

	if ((left_result == -1) || (right_result == -1))
		return;

	if ((left_result == 1) || (right_result == 1))
	{
		sprintf (command, "%s -b 142 -l %s -r %s -o %s", opt_mpo_command, jpg_left, jpg_right, mpo_path);

		if (opt_concise)
			fprintf (stderr, "%s + %s --> %s\n", jpg_left, jpg_right, mpo_path);
		else
			fprintf (stderr, "\t%s\n", command);

		result = system (command);
	//	fprintf (stderr, "\tresult = %d\n", result);
	}
}

/*-----------------------------------------------------------------------------
 *	NAME
 *		raw2jpg main routine
 *
 *	SYNOPSIS
 *		int
 *		main(
 *			int				argc,				- Number of Arguments
 *			char			*argv[])			- Argument Vector
 *
 *	RETURN VALUE
 *		None.
 *
 *	DESCRIPTION
 *		This function converts files from raw to jpg.
 *-----------------------------------------------------------------------------
 */

int main (int argc, const char *argv[], const char *envp[])
{
	int							opt_help = 0;
	int							opt_debug = 0;
	char						buffer [1024];
	char						working_directory [4096];
	int							i;
	const char					*cp;
	const char					*ep;

	typedef struct
	{
		const char * const	name;
		void				*value;
		const char * const	help;
	}
	option_structure;

	static option_structure option_list [] =
	{
		{ "-?",	&opt_help,				"Display these help messages" },
		{ "-h",	&opt_help,				"Display these help messages" },
		{ "-c",	&opt_concise,			"Concise messages" },
		{ "-d",	&opt_debug,				"Debug messages" },
		{ "=j",	&opt_jpg,				"JPEG Destination Folder" },
		{ "=lf",&opt_left,				"Left Source Folder" },
		{ "=lr",&opt_left_rotation,		"Left Rotation (Clockwise Degrees)" },
		{ "=rf",&opt_right,				"Right Source Folder" },
		{ "=rr",&opt_right_rotation,	"Right Rotation (Clockwise Degrees)" },
		{ "=r",	&opt_rotation,			"Non-Stereoscopic Rotation (Clockwise Degrees)" },
		{ "-rp",&opt_rotate_psd,		"Also Rotate PSD Files" },
		{ "=mc",&opt_mpo_command,		"MPO Command" },
		{ "=s",	&opt_suffix,			"JPG Filename Suffix" },
		{ "=mf",&opt_mpo,				"MPO Destination Folder" },
		{ "-v",	&opt_verbose,			"Verbose messages" },
	};
	int	number_of_options = sizeof (option_list) / sizeof (option_structure);


//	convert -rotate "90" in.jpg out.jpg

	/*-------------------------------------------------------------------------
	 *	Process Command Line and Options
	 *-------------------------------------------------------------------------
	 */

	cp = strrchr (argv[0], DIRECTORY);
	if (cp == (const char *) NULL)
		cp = argv[0];
	else
		cp++;

	ep = strrchr (cp, '.');
	if (ep == (const char *) NULL)
		ep = strchr (cp, '\0');

	strncpy (my_name, cp, ep - cp);
	my_name [ep - cp] = '\0';

	argv++;
	argc--;

	while ((argc > 0) && (**argv == OPTION))
	{
		for (i = 0; i < number_of_options; i++)
		{
			if (option_list[i].name[0] == '-')
			{
				if (strcmp (option_list [i].name + 1, *argv + 1) == 0)
				{
					*((int *) (option_list [i].value)) ^= 1;
					argv++;
					argc--;
					break;
				}
			}
			if (option_list[i].name[0] == '=')
			{
				if (strcmp (option_list [i].name + 1, *argv + 1) == 0)
				{
					if (argc > 1)
					{
						argv++;
						argc--;
						*((const char **) (option_list [i].value)) = *argv;
						argv++;
						argc--;
						break;
					}
					else
						fprintf (stderr, "Error: required value missing for %s\n", *argv);
				}
			}
		}

		/*---------------------------------------------------------------------
		 *	If the option was not found, try a case-insensitive search.
		 *---------------------------------------------------------------------
		 */

		if (i == number_of_options)
		{
			for (i = 0; i < number_of_options; i++)
			{
				if (option_list[i].name[0] == '-')
				{
					if (strcasecmp (option_list [i].name + 1, *argv + 1) == 0)
					{
						*((int *) (option_list [i].value)) ^= 1;
						argv++;
						argc--;
						break;
					}
				}
				if (option_list[i].name[0] == '=')
				{
					if (strcasecmp (option_list [i].name + 1, *argv + 1) == 0)
					{
						if (argc > 1)
						{
							argv++;
							argc--;
							*((const char **) (option_list [i].value)) = *argv;
							argv++;
							argc--;
							break;
						}
						else
							fprintf (stderr, "Error: required value missing for %s\n", *argv);
					}
				}
			}
		}

		if (i == number_of_options)
		{
			fprintf (stderr, "Error: unrecognized option %s\n", *argv);
			opt_help = 1;
			break;
		}
	}

	if (opt_help)
	{
		fprintf (stderr, "%s: Convert raw to jpg file (using ImageMagick)\n", my_name);
		fprintf (stderr, "Copyright (C) 2016-2018 Herb Weiner <herbw@wiskit.com>\n");
		fprintf (stderr, "http://www.herbweiner.com/stereo-software.html\n");
		fprintf (stderr, "Licensed under Creative Commons CC BY-SA 4.0\n\n");

		fprintf (stderr, "usage: %s -options\n", my_name);
		fprintf (stderr, "options:\n");
		for (i = 0; i < number_of_options; i++)
		{
			if (option_list[i].name[0] == '-')
				fprintf (stderr, "  %c%s: %s [%s]\n",
							OPTION, option_list [i].name + 1, option_list [i].help,
							*((int *) (option_list [i].value)) ? "enabled" : "disabled");
			else
				fprintf (stderr, "  %c%s value: %s [%s]\n",
							OPTION, option_list [i].name + 1, option_list [i].help,
							*((const char **) (option_list [i].value)) ? (const char *) *((const char **) (option_list [i].value)) : "<UNDEFINED>");
		}
	//	fprintf (stderr, "Press ENTER to continue . . .\n");
	//	fgets (buffer, sizeof (buffer), stdin);
		exit (1);
	}

	getcwd (working_directory, sizeof (working_directory));

	if (opt_debug)
	{
		fprintf (stderr, "%s: Working Directory: %s\n", my_name, working_directory);

		i = 0;
		while (envp [i] != NULL)
		{
			if (strncasecmp (envp [i], "PATH=", 5) == 0)
			{
				fprintf (stderr, "PATH:\n");
				ep = envp [i] + 5;
				while (*ep != '\0')
				{
					cp = ep;
					ep = strchr (cp, PATHSEP);
					if (ep == (const char *) NULL)
						ep = strchr (cp, '\0');
					if (ep > cp)
						fprintf (stderr, "    %.*s\n", (int) (ep - cp), cp);
					if (*ep != '\0')
						ep++;
				}
				fprintf (stderr, "\n");
				break;
			}
			i++;
		}
	}

	if (argc < 1)
	{
		/*---------------------------------------------------------------------
		 *	If you double click on the Windows .exe, it will open a Command
		 *	window that will close when the program terminates. Provide
		 *	a brief synopsis of the program, and wait for the user to Press
		 *	ENTER before closing the window.
		 *---------------------------------------------------------------------
		 */

		fprintf (stderr, "raw2jpg Copyright (c) 2018 Herb Weiner <herbw@wiskit.com>. All rights reserved.\n");
		fprintf (stderr, "Converts raw (or psd) files to jpg files\n");
		fprintf (stderr, "    using ImageMagick (https://www.imagemagick.org) convert command.\n");
		fprintf (stderr, "Use %s %ch for more information.\n", my_name, OPTION);

		if (WINDOWS)
		{
			fprintf (stderr, "Press ENTER to continue . . .\n");
			fgets (buffer, sizeof (buffer), stdin);
		}

		exit (0);
	}

	for (i = 0; i < argc; i++)
	{
		process_file (argv [i]);
	}
}
